/******************************************************************************
 * Copyright (c) 2006, Intalio Inc.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *     Intalio Inc. - initial API and implementation
 *******************************************************************************
 * Dates       		 Author              Changes
 * Jan 10, 2007      Antoine Toulm   Creation
 */
package org.eclipse.stp.bpmn.export;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.util.LinkedList;
import java.util.List;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.eclipse.gmf.runtime.common.core.util.Log;
import org.eclipse.gmf.runtime.common.ui.util.WindowUtil;
import org.eclipse.gmf.runtime.diagram.ui.image.ImageFileFormat;
import org.eclipse.gmf.runtime.diagram.ui.render.internal.DiagramUIRenderPlugin;
import org.eclipse.gmf.runtime.diagram.ui.render.internal.DiagramUIRenderStatusCodes;
import org.eclipse.gmf.runtime.diagram.ui.render.internal.l10n.DiagramUIRenderMessages;
import org.eclipse.gmf.runtime.diagram.ui.render.util.CopyToImageUtil;
import org.eclipse.gmf.runtime.notation.Diagram;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.wizard.Wizard;
import org.eclipse.osgi.util.NLS;
import org.eclipse.stp.bpmn.diagram.part.BpmnDiagramEditorPlugin;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTError;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Combo;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.DirectoryDialog;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.IExportWizard;
import org.eclipse.ui.IWorkbench;
import org.eclipse.ui.dialogs.WizardExportResourcesPage;

/**
 * This wizard is based on the work of Anthony Hunter and cmahoney in CopyToImageAction.
 * 
 * Instead of using a dialog, we use a wizard to do the job.
 * 
 * Note that the wizard can only export BPMN diagrams.
 * 
 * @author <a href="mailto:atoulme@intalio.com">Antoine Toulm</a>
 * @author <a href="http://www.intalio.com">&copy; Intalio, Inc.</a>
 */
public class ImageExportWizard extends Wizard implements IExportWizard {

	/**
	 * The page used to hold the destination fields. Does the essential of the work.
	 * 
	 * @author <a href="mailto:atoulme@intalio.com">Antoine Toulm</a>
	 * @author <a href="http://www.intalio.com">&copy; Intalio, Inc.</a>
	 */
	private class ExportImagePage extends WizardExportResourcesPage {

		private Text folderText;
		private boolean folderValid;
		private String folder = "";
		private boolean fileNameValid = false;
		private Text fileNameText;
		private String fileName = "image";
		private Combo imageFormatCombo;
		private ImageFileFormat imageFormat = ImageFileFormat.JPG;

		public ExportImagePage(IStructuredSelection sel) {
			super("Select the diagram to export", sel);
			setTitle("Export a BPMN diagram as an image");
			setDescription("Choose a file and a format to export the " +
					"diagram as an image");
			
		}

		/**
		 * utility method to create a composite widget
		 * @param parent the parent widget
		 * @param columns the number of columns in the grid layout for the new
		 * composite.
		 * @return the new composite widget
		 */
		private Composite createComposite(Composite parent, int columns) {
			Composite composite = new Composite(parent, SWT.NONE);

			GridLayout gridLayout = new GridLayout();
			gridLayout.numColumns = columns;
			gridLayout.marginHeight = 0;
			gridLayout.marginWidth = 0;
			gridLayout.makeColumnsEqualWidth = false;

			GridData data =
				new GridData(
					GridData.HORIZONTAL_ALIGN_FILL | GridData.FILL_HORIZONTAL);

			composite.setLayoutData(data);
			composite.setLayout(gridLayout);

			return composite;
		}

		/**
		 * Creates the Destination Group, the text representing the
		 * folder to export to and the file name for the image.
		 */
		@Override
		protected void createDestinationGroup(Composite parent) {
			Composite composite = createComposite(parent, 3);
			Label folderLabel = new Label(composite,SWT.NONE);
			folderLabel.setText("Export folder");
			folderText = new Text(composite, SWT.BORDER);
			folderText.setText(folder == null ? ""  : folder);
			folderText.addModifyListener(new ModifyListener() {
				public void modifyText(ModifyEvent e) {
					validateFolderText();
				}
			});
			GridData gridData = new GridData(GridData.FILL_HORIZONTAL);
			gridData.widthHint = 250;
			folderText.setLayoutData(gridData);

			Button button = new Button(composite, SWT.PUSH);
			button.setText("Browse");
			button.setLayoutData(WindowUtil.makeButtonData(button));
			button.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent event) {
					handleBrowseButtonPressed();
				}
			});
			Label fileNameLabel = new Label(composite,SWT.NONE);
			fileNameLabel.setText("Image file name");
			fileNameText = new Text(composite, SWT.BORDER);
			fileNameText.setText(fileName + "." + imageFormat.getName().toLowerCase());
			fileNameText.addModifyListener(new ModifyListener() {
				public void modifyText(ModifyEvent e) {
					validateFileNameText();
				}
			});
			gridData =
				new GridData(
					GridData.HORIZONTAL_ALIGN_FILL | GridData.GRAB_HORIZONTAL);
			gridData.widthHint = 250;
			fileNameText.setLayoutData(gridData);
		}

		/**
		 * not implemented.
		 */
		public void handleEvent(Event event) {
		}
		
		
		@SuppressWarnings("unchecked")
		@Override
		protected List getTypesToExport() {
			List types = new LinkedList();
			types.add("bpmn_diagram");
			return types;
		}
		
		@SuppressWarnings("unchecked")
		@Override
		public List getSelectedResources() {
			return super.getSelectedResources();
		}
		/**
		 * The options group contains the image fomat combo box.
		 */
		@Override
		protected void createOptionsGroupButtons(Group optionsGroup) {
			Composite composite = createComposite(optionsGroup, 2);
			Label label = new Label(composite,SWT.NONE);
			label.setText("Image format");
			imageFormatCombo = new Combo(composite, SWT.DROP_DOWN | SWT.READ_ONLY);
			imageFormatCombo.setItems(getImageFormatItems());
			imageFormatCombo.setText(imageFormat.getName());
			imageFormatCombo.addSelectionListener(new SelectionAdapter() {
				public void widgetSelected(SelectionEvent e) {
					imageFormat =
						ImageFileFormat.resolveImageFormat(
							imageFormatCombo.getSelectionIndex());
	                
	                // update filename to reflect new format
	                fileNameText.setText(fileName + "." + imageFormat.getName().toLowerCase());
	                validateFileNameText();
				}
			});
			GridData gridData =
				new GridData(
					GridData.HORIZONTAL_ALIGN_FILL | GridData.FILL_HORIZONTAL);
			gridData.widthHint = 250;
			imageFormatCombo.setLayoutData(gridData);
		}
		
		/**
		 * get the supported image formats from the enumerated type.
		 * @return array of supported image formats.
		 */
		private String[] getImageFormatItems() {
			String[] items = new String[ImageFileFormat.VALUES.length];
			for (int i = 0; i < ImageFileFormat.VALUES.length; i++) {
				items[i] = ImageFileFormat.VALUES[i].getName();
			}
			return items;
		}
		/**
		 * handle a browse button pressed selection.
		 */
		private void handleBrowseButtonPressed() {
			DirectoryDialog dialog =
				new DirectoryDialog(Display.getCurrent().getActiveShell());
			dialog.setMessage("Export image");
			dialog.setText("Choose a folder to hold the exported image");

			String dirName = folderText.getText();
			if (!dirName.equals("")) {
				File path = new File(dirName);
				if (path.exists())
					dialog.setFilterPath(new Path(dirName).toOSString());
			}

			String selectedDirectory = dialog.open();
			if (selectedDirectory != null) {
				folderText.setText(selectedDirectory);
			}
		}
		
		/**
		 * validate the folder text field.
		 */
		private void validateFolderText() {

			if (folderText.getText().equals("")) {
				setMessage(null);
				folderValid = false;
				return;
			}

			IPath path = new Path("");
			if (!path.isValidPath(folderText.getText())) {
				setErrorMessage("The folder is invalid");
				folderValid = false;
				return;
			}

			File file = new File(folderText.getText());
			if (!file.exists()) {
				setErrorMessage("The folder does not exist.");
				folderValid = false;
				return;
			}

			folderValid = true;
			folder = folderText.getText();
			if (fileNameValid) {
				setMessage(null);
			} else {
				validateFileNameText();
			}
		}
		
		/**
		 * validate the file name text field.
		 */
		private void validateFileNameText() {
			IStatus nameStatus =
				ResourcesPlugin.getWorkspace().validateName(
					fileNameText.getText(),
					IResource.FILE);

			if (!nameStatus.isOK()) {
				setErrorMessage(nameStatus.getMessage());
				fileNameValid = false;
				return;
			}

			fileNameValid = true;
	        
	        IPath filePath = (new Path(fileNameText.getText())).removeFileExtension();
			fileName = filePath.toString();
			if (folderValid) {
				setMessage(null);
			} else {
				validateFolderText();
			}
		}
		/**
		 * Computes the destination for the image file
		 * @return
		 */
		public IPath getDestinationPath() {
			IPath path = new Path(folder);
			String fnWithExtension = fileName +  "." + 
				imageFormat.getName().toLowerCase();
			
			return path.append(fnWithExtension);
		}

		/**
		 * @return the imageFormat
		 */
		public ImageFileFormat getImageFormat() {
			return imageFormat;
		}
		
	}
	
	private ExportImagePage imagePage;

	public ImageExportWizard() {
	}

	/**
	 * Does the same thing as the CopyToImageAction#run() method.
	 * 
	 */
	@SuppressWarnings("unchecked")
	@Override
	public boolean performFinish() {
		List exporting = imagePage.getSelectedResources();
		if (exporting.isEmpty()) {
			imagePage.setErrorMessage("No resources are selected");
			return false;
		}
		for (Object toExport : exporting) {
			if (toExport == null || !(toExport instanceof IFile)) {
				return false;
			}
			if (!((IFile) toExport).getFileExtension().equals("bpmn_diagram")) {
				imagePage.setErrorMessage("A file with an invalid extension is " +
						"selected. Deselect it and try again.");
				return false;
			}
		}

		// if more than one diagram is selected.
		if (exporting.size() > 1) {
			imagePage.setErrorMessage("Please select one diagram only");
			return false;
		}
		// now getting serious and exporting for real

		Object selected = imagePage.getSelectedResources().get(0);
		
		IRunnableWithProgress runnable = createRunnable((IFile) selected,
				imagePage.getDestinationPath(),
				imagePage.getImageFormat());

		ProgressMonitorDialog progressMonitorDialog = new ProgressMonitorDialog(
				Display.getCurrent().getActiveShell());
		try {
			progressMonitorDialog.run(false, true, runnable);
		} catch (InvocationTargetException e) {
			Log.warning(DiagramUIRenderPlugin.getInstance(),
					DiagramUIRenderStatusCodes.IGNORED_EXCEPTION_WARNING, e
					.getTargetException().getMessage(), e.getTargetException());

			if (e.getTargetException() instanceof OutOfMemoryError) {
				openErrorDialog(DiagramUIRenderMessages.CopyToImageAction_outOfMemoryMessage);
			} else if (e.getTargetException() instanceof SWTError) {
				/**
				 * SWT returns an out of handles error when processing large
				 * diagrams
				 */
				openErrorDialog(DiagramUIRenderMessages.CopyToImageAction_outOfMemoryMessage);
			} else {
				openErrorDialog(e.getTargetException().getMessage());
			}
			return true;
		} catch (InterruptedException e) {
			/* the user pressed cancel */
			Log.warning(DiagramUIRenderPlugin.getInstance(),
					DiagramUIRenderStatusCodes.IGNORED_EXCEPTION_WARNING, e
					.getMessage(), e);
		}


		return true;
	}

	/**
	 * display an error dialog
	 * 
	 * @param message
	 *            cause of the error
	 */
	private void openErrorDialog(String message) {
		MessageDialog
			.openError(
				Display.getCurrent().getActiveShell(),
				DiagramUIRenderMessages.CopyToImageAction_copyToImageErrorDialogTitle,
				NLS
					.bind(
						DiagramUIRenderMessages.CopyToImageAction_copyToImageErrorDialogMessage,
						message));
	}
	
	public void init(IWorkbench workbench, IStructuredSelection selection) {
		setWindowTitle("BPMN to Image Export Wizard");
		imagePage = new ExportImagePage(selection);
		addPage(imagePage);
	}

	/**
	 * copy the selected shapes in the diagram to an image file.
	 * 
	 * @param diagramEditPart
	 *            the diagram editor
	 * @param list
	 *            list of selected shapes in the diagram
	 * @param destination
	 *            path to the new image file
	 * @param imageFormat
	 *            image format to create
	 * @return the runnable with a progress monitor
	 */
	private IRunnableWithProgress createRunnable(final IFile diagram,final IPath destination,final ImageFileFormat format) {
		return new IRunnableWithProgress() {

			public void run(IProgressMonitor monitor) {
					ResourceSet resourceSet = new ResourceSetImpl();
					TransactionalEditingDomain.Factory.INSTANCE.createEditingDomain(resourceSet);
					Resource res = resourceSet.getResource(
							URI.createPlatformResourceURI(
									diagram.getFullPath().toString(),true), 
									true);
					Diagram tempD = null;

					for (Object resElt : res.getContents()) {
						if (resElt instanceof Diagram) {
							tempD= (Diagram) resElt;

						}
					}
					if (tempD == null) {
						return ;
					}
					final Diagram diagram = tempD;
					
					try {
						new CopyToImageUtil().copyToImage(diagram, destination, 
								format, new NullProgressMonitor(), 
								BpmnDiagramEditorPlugin.DIAGRAM_PREFERENCES_HINT);
					} catch (CoreException e) {
						BpmnDiagramEditorPlugin.getInstance().getLog().log(
								new Status(
										IStatus.ERROR,
										BpmnDiagramEditorPlugin.ID,
										IStatus.ERROR,e.getMessage(),e));
					}
			}
		};
	}
	
	
	
}
